//----------------------------------------------------------------------
// Prozessor : ATmega32
// Sprache   : C
// Datum     : 23.06.2011
// Version   : 0.9
// Autor     : Thilo M.
// Programm  : Motoransteuerung 12V Vorwrts - Rckwarts mit PWM
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// Variablen
//----------------------------------------------------------------------

double X_min = 90;		// Maximalweg der Mechanik
double X_max = 270;
double Y_min = 16;		// Mindesthhe am 21.12.
double Y_max = 65;		// Maximalhhe am 21.06.
double ADCX_Null = 103;			// diese Werte mssen empirisch hardwarebezogen ermittelt werden
double ADCX_Ende = 923;
double ADCY_Null = 0;
double ADCY_Ende = 1023;
double X_Schrittweite = 0;
double Y_Schrittweite = 0;
double X_Messwert = 0;
double Y_Messwert = 0;
// Variablen fr Dmpfungen der Messwerte
double ZKX = 1.4;		// Zeitkonstante fr Dmpfung
double ZKY = 1.4;		// Zeitkonstante fr Dmpfung
double TGX = 1;			// Gemittelter Wert X
double TGY = 1;			// Gemittelter Wert Y

//----------------------------------------------------------------------
// Definition der Ausgnge
//----------------------------------------------------------------------

#define PWMPORT		PORTB	// MOSFET fr PWM (Sanftanlauf)
#define PWMDDR		DDRB
#define PWMPIN		3
#define R_LPORT		PORTB	// Relais Rechts- Linkslauf
#define R_LDDR		DDRB
#define R_LPIN		2
#define X_YPORT		PORTB	// Relais X- Y-Motor Umschaltung
#define X_YDDR		DDRB
#define X_YPIN		1

#define Ansprechschwelle_X	1	// alle wieviel  soll nachgefhrt werden
#define Ansprechschwelle_Y	1

//----------------------------------------------------------------------
// Inititialisierung Timer und Ausgnge
//----------------------------------------------------------------------

void init_Motorcontrol (void)
{
	PWMDDR |= (1<<PWMPIN);		// Pin als Ausgang
	PWMPORT &= ~(1<<PWMPIN);	// Pin auf '0'
	R_LDDR |= (1<<R_LPIN);		// Pin als Ausgang
	R_LPORT &= ~(1<<R_LPIN);	// Pin auf '0'
	X_YDDR |= (1<<X_YPIN);		// Pin als Ausgang
	X_YPORT &= ~(1<<X_YPIN);	// Pin auf '0'
	
// Einstellungen fr Timer 0 (3.9kHz-PWM fr Sanftanlauf)
	
	TCCR0 |= _BV(WGM00)|_BV(COM01)|_BV(CS01); //Phase-correct PWM
	OCR0 = 0;	// Ausgang AUS
	
	X_Schrittweite = (X_max-X_min)/(ADCX_Ende-ADCX_Null);
	Y_Schrittweite = (Y_max-Y_min)/(ADCY_Ende-ADCY_Null);
}

//----------------------------------------------------------------------
// Motor X einschalten (Sanftanlauf)
//----------------------------------------------------------------------//----------------------------------------------------------------------
// Messung der Positionen und Nachfhrung
//----------------------------------------------------------------------

void Messung (void)
{
	Status &= ~(1<<ST_Messung);
	sun(dt, dz, laenge / KO, breite / KO);	
	double X_Soll = azi * KO;
	double Y_Soll = elev * KO;
	X_Messwert = ((ADC_conversion(0) - ADCX_Null) * X_Schrittweite) + X_min;
	Y_Messwert = ((ADC_conversion(1) - ADCY_Null) * Y_Schrittweite) + Y_min;
	
	TGX = TGX*(1-1/ZKX)+X_Messwert/ZKX;	// Dmpfung X
	X_Messwert = TGX;
	TGY = TGY*(1-1/ZKY)+Y_Messwert/ZKY;	// Dmpfung Y
	Y_Messwert = TGY;
	
	double DX = X_Messwert - X_Soll;
	double DY = Y_Messwert - Y_Soll;
}

// hier Code fr Ansteuerung der Hardware einfgen