//***************************************************************************************
// Definition des DCF-Protokolles
//---------------------------------------------------------------------------------------
// Bit  0..14: Wetterinformationen
// Bit 15    : 1=> Reserveantenne aktiv
// Bit 16    : 1=> Umstellung Sommer/Winterzeit erfolgt in der nchsten Minute
// Bit 17    : 1=> Sommerzeit aktiv	(Bit 18 = 0)
// Bit 18    : 1=> Normalzeit aktiv (Bit 17 = 0)
// Bit 19    : 1=> Schaltsekunde wird ab der nchsten Minute bercksichtigt
// Bit 20	 : Startbit, immer 1
// Bit 21..27: Minuten
// Bit 28    : Parittsbit 1
// Bit 29..34: Stunden
// Bit 35    : Parittsbit 2
// Bit 36..41: Kalendertag
// Bit 42..44: Wochentag
// Bit 45..49: Kalendermonat
// Bit 50..57: Kalenderjahr
// Bit 58    : Parittsbit 3
// Bit 59    : fehlt
// Parittsbits: gerade Paritt, Summe der Bits muss gerade sein
//---------------------------------------------------------------------------------------
//***************************************************************************************

volatile unsigned const char DCF_BITS[59] = {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,3,2,1,1,1,2,4,8,10,20,40,0,1,2,4,8,10,20,0,1,2,4,8,10,20,1,2,4,1,2,4,8,10,1,2,4,8,10,20,40,80,0};
// statusbyte (Flags)
volatile const char Uhr_DCF_BIT = 0;			// '0' => noch nicht synchronisiert, 1=> Synchronisiert
volatile const char Uhr_SC = 1;					// '1' => Uhr mit nchster Sekunde stellen
volatile const char Uhr_Syncfreigabe = 2;		// '1' => Uhr auf DCF77 synchronisieren
volatile const char Uhr_nicht_gestellt = 3;		// '1' => Uhr nicht gestellt, keine Funktion freigeben

volatile unsigned char UhrStatus = 0;
volatile unsigned char DCF_Minute = 0;			// 
volatile unsigned char DCF_Stunde = 0;			// 
volatile unsigned char DCF_Tag = 0;				// 
volatile unsigned char DCF_WTag = 0;				// 
volatile unsigned char DCF_Monat = 0;			// 
volatile unsigned int DCF_Jahr = 0;		
volatile unsigned char DCF_MESZ_cnt = 0;
volatile unsigned char parcnt = 0;				// zhler fr Parittsprfung
// RTC-Variablen
/*
volatile unsigned char Sekunde = 0;			// 
volatile unsigned char Minute = 59;			// 
volatile unsigned char Stunde = 5;			// 
volatile unsigned char Tag = 17;				// 
volatile unsigned char WTagNr = 0;				// 
volatile unsigned char WTag1 = 0;
volatile unsigned char WTag2 = 0;
volatile unsigned char Monat = 6;			// 
volatile int Jahr = 2011;						// 
*/
volatile int mod_Jahr = 0;						// Nachkommaanteil fr Schaltjahrberechnung
volatile unsigned char Monatstage[13] = {0,32,29,32,31,32,31,32,32,31,32,31,32};
// DCF77 - Variablen
volatile unsigned int ibr_cnt = 0;		// Impulsbreitenmessung
volatile unsigned int ips_cnt = 0;		// Impulspausenmessung
volatile unsigned char DCF_bitcnt = 0;	// Sekundenzhler
//float Sommerzeit = 0;

volatile unsigned char Zykl_cnt = 1;
volatile unsigned char Mess_cnt = 0;


//***************************************************************************************
//***************************************************************************************

void init_clock (void)
{
// Pin fr DCF77 - Eingangssignal

	DCF_DDR &= ~(1<<DCF_PIN);	// Pin auf Eingang
	DCF_PORT |= (1<<DCF_PIN);	// Pullup EIN
	
// Einstellungen fr Timer 1

	TCCR1A = 0b00000000;
	TCCR1B |= _BV(WGM12)|_BV(CS12)|_BV(CS10);	// Prescaler /1024, CTC - Mode 
	TCNT1 = 0;									// Timer rcksetzen
	OCR1A = 15624;								// OCR1A = 15625 (16MHz / 1024 / 15625 = 1s)
	
// Einstellungen fr Timer 2 (10ms-Timer fr DCF-Abtastung)
	
	TCCR2 |= _BV(CS20)|_BV(CS21)|_BV(CS22); //Teiler auf CPU-Takt/1024 setzen
	TCNT2 = 99;	// 10ms zum Starten
	
// Timerflags lschen und Interrupts freigeben
	
	TIFR |= _BV(OCF1A)|_BV(TOV2)|_BV(OCF0);		// Flags lschen
	TIMSK |= _BV(OCIE1A)|_BV(TOIE2);				// Timer starten
	
	UhrStatus &= ~_BV(Uhr_DCF_BIT);			// nicht Synchronisiert
	UhrStatus |= _BV(Uhr_Syncfreigabe);			// Synchronisieren freigeben
	UhrStatus |= _BV(Uhr_nicht_gestellt);			// Funktionen sperren, wenn Uhr nicht gestellt ist
}

//***************************************************************************************


//***************************************************************************************

void clr_DCF (void)		// Protokoll rcksetzen
{
	DCF_Minute = 0;			// 
	DCF_Stunde = 0;			// 
	DCF_Tag = 0;			// 
	DCF_WTag = 0;			// 
	DCF_Monat = 0;			// 
	DCF_Jahr = 0;			//
	DCF_MESZ_cnt = 0;
	parcnt = 0;				// Parittsbit - Zhler rcksetzen
	DCF_bitcnt = 0;
}

//***************************************************************************************

void check_parity (void)		// Parittsprfung der Prfbits
{
	if (!(parcnt%2 == 0))	// falls Nachkommaanteil bei Division /2 vorhanden, dann
		{
			clr_DCF();					// DCF-Uhrzeit rcksetzen fr neue Messung
			DCF_bitcnt = 0;				// DCF - Bitzhler lschen
			UhrStatus &= ~_BV(Uhr_DCF_BIT); // DCF nicht Synchronisiert
		}
	parcnt = 0;			// Parittsbit - Zhler rcksetzen
}

//***************************************************************************************

void set_clock (void)		// Bei korrekt empfangenem DCF-Protokoll Uhr stellen
{
	TCNT1 = 0;						// RTC Synchronisieren
	TIFR |= _BV(OCF1A);				// Flag lschen
	TCCR1B |= _BV(CS12)|_BV(CS10); // Timer starten
	Sekunde = 0;					// 
	Minute = DCF_Minute;			// 
	Stunde = DCF_Stunde;			// 
	Tag = DCF_Tag;					// 
	WTagNr = DCF_WTag-1;
	Monat = DCF_Monat;				// 
	Jahr = 2000+DCF_Jahr;			// Jahrtausend dazuaddieren
	if (DCF_MESZ_cnt == 3)
		{
			Sommerzeit = 1;
		}
	if (DCF_MESZ_cnt == 2)
		{
			Sommerzeit = 0;
		}				
	UhrStatus &= ~_BV(Uhr_SC);		// DCF - Synchronflag setzen und Zyklenzhler
	UhrStatus |= _BV(Uhr_DCF_BIT);					// von Hand gestellt
	clr_DCF ();										// auf DCF. DCF-Uhrzeit rcksetzen fr neue Messung
	UhrStatus &= ~_BV(Uhr_Syncfreigabe);			// Synchronisieren sperren
	UhrStatus &= ~_BV(Uhr_nicht_gestellt);			// Funktionen freigeben
	fprintf(stdout, "\n");
}

//***************************************************************************************

void DCF_Zuweisen (void)		// Wertigkeit des jeweiligen DCF-Bits zuweisen
{								// (nur wenn '1' empfangen wurde)
if ((DCF_bitcnt >= 17)&&(DCF_bitcnt <= 18))
	{
		DCF_MESZ_cnt += DCF_BITS[DCF_bitcnt];
	}
if ((DCF_bitcnt >= 21)&&(DCF_bitcnt <= 27))
	{
		DCF_Minute += DCF_BITS[DCF_bitcnt];
	}
if ((DCF_bitcnt == 28)||(DCF_bitcnt == 35)||(DCF_bitcnt == 58))
	{
		parcnt ++;		// Parittsbitzhler erhhen
	}
if ((DCF_bitcnt >= 29)&&(DCF_bitcnt <= 34))
	{
		DCF_Stunde += DCF_BITS[DCF_bitcnt];
	}
if ((DCF_bitcnt >= 36)&&(DCF_bitcnt <= 41))
	{
		DCF_Tag += DCF_BITS[DCF_bitcnt];
	}
if ((DCF_bitcnt >= 42)&&(DCF_bitcnt <= 44))
	{
		DCF_WTag += DCF_BITS[DCF_bitcnt];
	}	
if ((DCF_bitcnt >= 45)&&(DCF_bitcnt <= 49))
	{
		DCF_Monat += DCF_BITS[DCF_bitcnt];
	}
if ((DCF_bitcnt >= 50)&&(DCF_bitcnt <= 57))
	{
		DCF_Jahr += DCF_BITS[DCF_bitcnt];
	}
}

//***************************************************************************************

//***************************************************************************************

ISR (SIG_OUTPUT_COMPARE1A) 	// RealTimeClock (Sekundenzhler)
{	
	Sekunde ++;
	if (Sekunde >= 60)
		{
			Minute ++;			// Minuten hochzhlen
			Sekunde = 0;
		}
	if (Minute >= 60)
		{
			Stunde ++;						// Stunden hochzhlen
			Minute = 0;
			UhrStatus &= ~_BV(Uhr_DCF_BIT);			// nicht Synchronisiert
			UhrStatus |= _BV(Uhr_Syncfreigabe);			// DCF77 Synchronisation freigeben
		}
	if (Stunde >= 24)
		{
			Tag ++;							// Tage hochzhlen
			WTagNr ++;
			if (WTagNr > 6)
			{
				WTagNr = 0;
			}
			Stunde = 0;
		}
	if (Monat == 2)						// im Februar auf Schaltjahr prfen
		{
			if ((((mod_Jahr = Jahr % 4) == 0)&&(!((mod_Jahr = Jahr % 100)== 0)))||((mod_Jahr = Jahr % 400) == 0))
				{
					Monatstage[2] = 30;		// Schaltjahr = Jahr teilbar durch 4 UND
				}							// NICHT teilbar durch 100 ODER durch 400
		}
	if (Tag >= Monatstage[Monat])
		{
			Monat ++;						// Monate hochzhlen
			Tag = 1;
		}
	if (Monat >= 13)
		{
			Jahr ++;						// Jahre hochzhlen
			Monat = 1;
		}
//---------------------------------------------------------------------------------------
	Status |= (1<<ST_Ausgabe);	// Displayausgabe freigeben
}

/***************************************************************************************/


/***************************************************************************************/

void check_DFbit (void) // Impulslngenauswertung mittels 10ms-Polling
{
	if ((!(DCF_INPORT &(1<<DCF_PIN))))	// wenn DF-Impuls 'LOW'
		{
			ibr_cnt++;	// Impulsbreitenzhler inkrementieren
			ips_cnt = 0;
//			PORTD &= ~_BV(LED_Betrieb);
			return;		// weiter bis Pin 'HIGH' ist
		}
//	PORTD |= _BV(LED_Betrieb);
	ips_cnt++;	// Impulspausenzhler inkrementieren
	if ((ibr_cnt >= 5)&&(ibr_cnt <= 10))	// wenn '0' erkannt
		{
			if ((DCF_bitcnt == 0)&&(UhrStatus &(1<<Uhr_SC)))	// mit Freigabe nach komplettem DCF - Protokoll
				{
					set_clock();								// Uhr am Anfang von Sekunde 0 stellen
				}
			if ((DCF_bitcnt == 28)||(DCF_bitcnt == 35)||(DCF_bitcnt == 58))
				{
					check_parity ();	// Prfbits auf gerade Paritt prfen
				}
			if ((DCF_bitcnt == 29)||(DCF_bitcnt == 36))
				{
					parcnt = 0;			// Anfang fr Parittsbitzhler bit 29 und 36
				}
			if (DCF_bitcnt == 58)	// wenn Paritt nach bit 58 OK, dann ist DCF synchronisiert
				{
					UhrStatus |= _BV(Uhr_DCF_BIT)|_BV(Uhr_SC);	// Synchronisiert
				}
			DCF_bitcnt ++;		// DCF - bits hochzhlen
			ibr_cnt = 0;
			return;
		}
	if ((ibr_cnt >= 16)&&(ibr_cnt <= 21))	// wenn '1' erkannt
		{
			DCF_Zuweisen ();		// bei '1' => Bitwertigkeit zuweisen
			if ((DCF_bitcnt == 0)&&(UhrStatus &(1<<Uhr_SC)))	// mit Freigabe nach komplettem DCF - Protokoll
				{
					set_clock();								// Uhr am Anfang von Sekunde 0 stellen
				}
			if ((DCF_bitcnt == 28)||(DCF_bitcnt == 35)||(DCF_bitcnt == 58))
				{
					check_parity ();		// Prfbits auf gerade Paritt prfen
				}
			if ((DCF_bitcnt == 29)||(DCF_bitcnt == 36))
				{
					parcnt = 0;				// Anfang fr Parittsbitzhler bit 29 und 36
				}
			if (DCF_bitcnt == 58)		// wenn Paritt nach bit 58 OK, dann ist DCF synchronisiert
				{
					UhrStatus |= _BV(Uhr_DCF_BIT)|_BV(Uhr_SC);	// Synchronisiert
				}
			if (DCF_bitcnt >= 21)		// erst ab bit 21 parittsbitzhler erhhen
				{
					parcnt ++;
				}
			DCF_bitcnt ++;			// DCF - bits hochzhlen
			ibr_cnt = 0;
			return;
		}
	if	(ips_cnt >= 150)			// >= 340ms: Timeout (kein Impuls empfangen)
		{
			if (DCF_bitcnt < 59)
			{
				clr_DCF ();	
			}
			DCF_bitcnt = 0;
			ibr_cnt = 0;
			ips_cnt = 100;
			UhrStatus &= ~_BV(Uhr_DCF_BIT); // DCF nicht Synchronisiert
		}
}

//----------------------------------------------------------------------
// Timer 2 Interrupt Service Routine (ISR)
//----------------------------------------------------------------------

ISR(TIMER2_OVF_vect)
{
	TCNT2 = 99; //Timer neu setzen fr nchste 10ms
	//------------------------------------------------------------------
	if (UhrStatus &(1<<Uhr_Syncfreigabe))
		{
			check_DFbit();	// DCF77 - Uhr
		}
	//------------------------------------------------------------------
	if (timeout_cnt < 3)
		{
			timeout_cnt ++;
		}
	else
		{
			uart_rx_cnt=0;
			uart_rx_buffer[0] = 0;
		}
	//------------------------------------------------------------------
	Mess_cnt ++;
	if (Mess_cnt > 33)
		{
			Status |= (1<<ST_Messung);
			Mess_cnt = 0;
		}
}



