/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <joerg@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.        Joerg Wunsch
 * ----------------------------------------------------------------------------
 *
 * Stdio demo, UART implementation
 *
 * $Id: uart.c,v 1.1.2.1 2005/12/28 22:35:08 joerg_wunsch Exp $
 */

#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include <avr/io.h>

#include "uart.h"

//------------------------------------------------------------------------------------------------
// UART - Variablen
volatile unsigned char timeout_cnt = 3;		// Zeitberlauf beim Empfang abfragen
volatile unsigned char uart_rx_cnt = 0;     // Zhler fr empfangene Zeichen
char uart_rx_buffer[13];      				// Empfangspuffer
volatile unsigned char UART_Befehl = 0;
volatile unsigned char UART_Tag = 0;
volatile unsigned char UART_Monat = 0;
volatile unsigned int UART_Jahr = 0;
volatile unsigned char UART_Stunde = 0;
volatile unsigned char UART_Minute = 0;
volatile unsigned char UART_Sekunde = 0;
volatile unsigned char UART_MESZ = 0;
volatile int8_t UART_GMT = 0;
float UART_Breite = 0;
float UART_Laenge = 0;
//------------------------------------------------------------------------------------------------



/*
 * Initialize the UART 
 */
void uart_init(void)
{
#if F_CPU < 2000000UL && defined(U2X)
  UCSRA = _BV(U2X);             // improve baud rate error by using 2x clk 
  UBRRL = (F_CPU / (8UL * UART_BAUD)) - 1;
#else
  UBRRL = (F_CPU / (16UL * UART_BAUD)) - 1;
#endif

  UCSRB = _BV(TXEN) | _BV(RXEN)| _BV(RXCIE); /* tx/rx enable */
}

/*
 * Send character c down the UART Tx, wait until tx holding register
 * is empty.
 */
int uart_putchar(char c, FILE *stream)
{

  if (c == '\a')
    {
      fputs("*ring*\n", stderr);
      return 0;
    }

  if (c == '\n')
    uart_putchar('\r', stream);
  loop_until_bit_is_set(UCSRA, UDRE);
  UDR = c;

  return 0;
}


int uart_getchar(FILE *stream)
{
  uint8_t c;
  char *cp, *cp2;
  static char b[RX_BUFSIZE];
  static char *rxp;

  if (rxp == 0)
    for (cp = b;;)
      {
	loop_until_bit_is_set(UCSRA, RXC);
	if (UCSRA & _BV(FE))
	  return _FDEV_EOF;
	if (UCSRA & _BV(DOR))
	  return _FDEV_ERR;
	c = UDR;
	/* behaviour similar to Unix stty ICRNL */
	if (c == '\r')
	  c = '\n';
	if (c == '\n')
	  {
	    *cp = c;
	    uart_putchar(c, stream);
	    rxp = b;
	    break;
	  }
	else if (c == '\t')
	  c = ' ';

	if ((c >= (uint8_t)' ' && c <= (uint8_t)'\x7e') ||
	    c >= (uint8_t)'\xa0')
	  {
	    if (cp == b + RX_BUFSIZE - 1)
	      uart_putchar('\a', stream);
	    else
	      {
		*cp++ = c;
		uart_putchar(c, stream);
	      }
	    continue;
	  }

	switch (c)
	  {
	  case 'c' & 0x1f:
	    return -1;

	  case '\b':
	  case '\x7f':
	    if (cp > b)
	      {
		uart_putchar('\b', stream);
		uart_putchar(' ', stream);
		uart_putchar('\b', stream);
		cp--;
	      }
	    break;

	  case 'r' & 0x1f:
	    uart_putchar('\r', stream);
	    for (cp2 = b; cp2 < cp; cp2++)
	      uart_putchar(*cp2, stream);
	    break;

	  case 'u' & 0x1f:
	    while (cp > b)
	      {
		uart_putchar('\b', stream);
		uart_putchar(' ', stream);
		uart_putchar('\b', stream);
		cp--;
	      }
	    break;

	  case 'w' & 0x1f:
	    while (cp > b && cp[-1] != ' ')
	      {
		uart_putchar('\b', stream);
		uart_putchar(' ', stream);
		uart_putchar('\b', stream);
		cp--;
	      }
	    break;
	  }
      }

  c = *rxp++;
  if (c == '\n')
    rxp = 0;

  return c;
}

//*********************************************************************************************

void check_RX (void)
{
	char String[10] = {0,0,0,0,0,0,0,0,0,0};
	
	for (unsigned char i=1;i<(strlen(uart_rx_buffer));i++)
		{
			String[i-1]=uart_rx_buffer[i];
		}

	switch (uart_rx_buffer[0])
		{
			case 'S':	UART_Sekunde = atoi(String);
						if ((UART_Sekunde >= 0)&&(UART_Sekunde <= 59))
							{
								TCNT1 = 0;						// RTC Synchronisieren
								TIFR |= _BV(OCF1A);				// Flag lschen
								TCCR1B |= _BV(CS12)|_BV(CS10); // Timer starten
								Sekunde = UART_Sekunde;
							}
						break;
			case 's':	UART_Stunde = atoi(String);
						if ((UART_Stunde >= 0)&&(UART_Stunde <= 23))
							{
								Stunde  = UART_Stunde;
							}
						break;
			case 'm':	UART_Minute = atoi(String);
						if ((UART_Minute >= 0)&&(UART_Minute <= 59))
							{
								Minute  = UART_Minute;
							}
						break;
			case 't':	UART_Tag = atoi(String);
						if ((UART_Tag >= 1)&&(UART_Tag <= 31))
							{
								Tag  = UART_Tag;
							}
						break;
			case 'M':	UART_Monat = atoi(String);
						if ((UART_Monat >= 1)&&(UART_Monat <= 12))
							{
								Monat  = UART_Monat;
							}
						break;
			case 'j':	UART_Jahr = atoi(String);
						if ((UART_Jahr >= 0)&&(UART_Jahr <= 9999))
							{
								Jahr  = UART_Jahr;
							}
						break;
			case 'g':	UART_GMT = atoi(String);
						if ((UART_GMT >= -12)&&(UART_GMT <= 12))
							{
								GMT = UART_GMT;
							}
						break;
			case 'z':	UART_MESZ = atoi(String);
						if ((UART_MESZ >= 0)&&(UART_MESZ <= 1))
							{
								Sommerzeit = UART_MESZ;
							}
						break;
			case 'l':	UART_Laenge = atof(String);
						if ((UART_Laenge >= 0)&&(UART_Laenge <= 360))
							{
								laenge = UART_Laenge;
							}
						break;
			case 'b':	UART_Breite = atof(String);
						if ((UART_Breite >= 0)&&(UART_Breite <= 360))
							{
								breite = UART_Breite;
							}
						break;
		}
	uart_rx_buffer[0] = 0;
}

//*********************************************************************************************
//*********************************************************************************************

ISR (USART_RXC_vect)			// RX Interrupt-complete-Routine
{
   timeout_cnt = 0;		// Timeout-Zhler rcksetzen
	
    uart_rx_buffer[uart_rx_cnt] = UDR;
	
   
    
        // ja, ist Ende des Strings (RETURN) erreicht?
        if ((uart_rx_buffer[uart_rx_cnt]=='\r')||(uart_rx_buffer[uart_rx_cnt]=='\n')) 
		{
			uart_rx_buffer[uart_rx_cnt] = 0;
            uart_rx_cnt=0;
			check_RX ();
			return;
        }
    uart_rx_cnt ++;   
}
